import os, sys
import gtk, gtk.glade, gobject, scipy, webbrowser
import cPickle
from StringIO import StringIO
from system import logger, dialogs, dataset, project
from system.octpy import matfile
import Numeric
import rpy
DATADIR = os.path.dirname(sys.modules['system'].__file__)
GLADEFILENAME = os.path.join(DATADIR, 'scicraft.glade')
                
        
class ExcelWriteExample:
    def run(self):
        # startup
        if not ensure_excel_available():
            return
            
        colids = ["Gnarfs","Foos","Moose"]
        rowids = ["Norway","Sweden","Argentina","Russia","Elbonia"]
        data   = scipy.array([[105,7,350],[75,9,310],[0,250,7],[9520,200,5202],[0,0,7]])
        data   = dataset.Dataset(data,[("Country",rowids),("Animal",colids)],name="TestDataset")
        writer = ExcelWriter("","")
        writer.run(data)        
                                                           
def ensure_excel_available(show_warnings = True):
    import sys
    if sys.platform != "win32":
        if show_warnings:
            logger.log("warning","Function requires Microsoft Excel installed with COM support. This is only availabe on Microsoft Windows operating systems")
        return False
    try:
        from win32com.client import Dispatch
    except Exception:
        if show_warnings:
            logger.log("Error","Error importing win32com library")
        return False
    try:                
        Dispatch("Excel.Application")
    except Exception:
        if show_warnings:    
            logger.log("warning","Error starting Microsoft Excel. Not installed?")
        return False
    return True

class ExcelReader:
    def insert_into(self, proj):
        return_data = self.run()
        proj.add_data(None,[return_data],fun="Excel Reader")
        
    def run(self):
        return_data = None
        import os
        if not ensure_excel_available():
            return

        ### Glade stuff
        try:
            WORKFLOWDIR = os.path.split(__file__)[0]
            xml = gtk.glade.XML(os.path.join(WORKFLOWDIR, "ExcelUI.glade"),
                            'dialog1')
            window = xml.get_widget('dialog1')        

            dispatch_excel_button = xml.get_widget('OpenExcel')
            dispatch_excel_label  = xml.get_widget('label5')
            read_sheet_button = xml.get_widget('button3')
            read_sheet_label  = xml.get_widget('label11')
            read_selected_button = xml.get_widget('button4')
            read_selected_label  = xml.get_widget('label12')

            dataset_name_input = xml.get_widget('entry1')
            dataset_name_label = xml.get_widget('label7')
            row_name_input = xml.get_widget('entry2')
            row_name_label = xml.get_widget('label8')
            column_name_input = xml.get_widget('entry3')
            column_name_label = xml.get_widget('label9')

            dimension_set_label1 = xml.get_widget('label13')
            dimension_set_label2 = xml.get_widget('label14')


            step1 = [read_sheet_button,read_sheet_label,read_selected_button,
                         read_selected_label]
            step2 = [dataset_name_input,dataset_name_label,
                         row_name_input,row_name_label,column_name_input,column_name_label,
                         dimension_set_label1,dimension_set_label2]

            self.disable_all(step1)
            self.disable_all(step2)
            
            dispatch_excel_button.connect("clicked",self.dispatch_excel,dispatch_excel_label,step1)
            read_selected_button.connect("clicked",self.read_selected,dispatch_excel_label,step1,step2)
            read_sheet_button.connect("clicked",self.read_sheet,dispatch_excel_label,step1,step2)
            run = window.run()

            if self.return_data and run == gtk.RESPONSE_OK:
                if dataset_name_input.get_text().strip():
                    self.return_data._name = dataset_name_input.get_text()
                if row_name_input.get_text().strip():
                    self.return_data.change_dim_name(self.return_data.get_dim_names()[0],row_name_input.get_text())
                if column_name_input.get_text().strip():
                    self.return_data.change_dim_name(self.return_data.get_dim_names()[1],column_name_input.get_text())                                               
        finally:
            window.destroy()
        return self.return_data            

    def use_rows_toggled(self,ignored):
        pass

    def use_cols_toggled(self,ignored):
        pass
            
    def dispatch_excel(self,ignored,label,step1):
        from win32com.client import Dispatch
        self.excel = Dispatch("Excel.Application")
        self.enable_all(step1)
        self.excel.Visible = 1
        label.set_label("Excel Active")

    def read_selected(self,ignored, label,step1,step2):

        try:
            read = self.excel.Selection.Value
            dims = len(read), len(read[0])
            dataname = self.excel.ActiveSheet.Name

            self.enable_all(step2)
            step2[0].set_text(dataname)
            step2[-1].set_label(str(dims))
            self.return_data = dataset.Dataset(read,name=dataname)
                          
        except Exception: 
            logger.log("warning","Error reading selected data from Excel")
            return

    def read_sheet(self,ignored,label,step1, step2):
        try:
            dataname = self.excel.Cells(3,1).Value
            rowname  = self.excel.Cells(5,1).Value
            colname =  self.excel.Cells(3,3).Value

            # find dimensions
            x,y = 5,3
            while(self.excel.Cells(x+1,2).Value):
                x += 1
            while(self.excel.Cells(4,y+1).Value):
                y += 1

            sheet = self.excel.ActiveSheet

            exlrange = sheet.Range(sheet.Cells(5,3),sheet.Cells(x,y))
            data = exlrange.Value 

            exlrange = sheet.Range(sheet.Cells(5,2),sheet.Cells(x,2))
            row_names = [name[0] for name in exlrange.Value]
        
            exlrange = sheet.Range(sheet.Cells(4,3),sheet.Cells(4,y))
            col_names = exlrange.Value[0]

            self.enable_all(step2[:-2])
            self.disable_all(step2[-2:])

            step2[0].set_text(dataname)
            step2[-1].set_label(str((x-4,y-2)))
            step2[2].set_text(rowname)
            step2[4].set_text(colname)
           
            self.return_data = dataset.Dataset(scipy.array(data),[(rowname,row_names),(colname,col_names)],name=dataname)

        except Exception:
            # something went wrong, urg!
            logger.log("warning","Error reading from Excel")

    def disable_all(self, li):
        for item in li:
            item.set_sensitive(False)
            
    def enable_all(self, li):
        for item in li:
            item.set_sensitive(True)              

class ExcelWriter:
    def run(self,data,*moredata):
        
        # startup
        if not ensure_excel_available():
            return
        from win32com.client import Dispatch
        excel = Dispatch("Excel.Application")

        # if no open workbooks, add a new one
        if not excel.ActiveWorkbook:
            excel.Workbooks.Add()
        try:
            # now add each dataset to a seperate sheet
            datas = list(moredata) + [data]
            for data in datas:
                            
                if len(data.shape)>2:
                    logger.log("notice","Ignoring dataset %s, too many dimensions" % (data.get_name(),))
                    continue
                if data.shape[1]>253 and data.shape[0]>253:
                    logger.log("notice","Ignoring dataset %s. Microsoft Excel may maximum have 256 columns." % (data.get_name(),))
                    continue
                if data.shape[1]>253 and data.shape[0]<254:
                    result = dialogs.get_yes_no("Transpose Dataset","Microsoft Excel may maximum have 256 columns.\n\n"
                              +"Dataset will be transposed.\n Is this OK?",default_response=gtk.RESPONSE_YES,
                              justify=gtk.JUSTIFY_CENTER)
                    if result:
                        new_array = Numeric.transpose(data.asarray())
                        dim_ids = [(dim_, data.get_identifiers(dim_)) for dim_ in data.get_dim_names()]
                        dim_ids.reverse()
                        data =dataset.Dataset(new_array, dim_ids,name="T("+data.get_name()+")")
                    else:
                      continue
                      
                sheet = excel.ActiveWorkbook.Sheets.Add()
                                
                try:
                    sheet.name = data.get_name()
                except Exception:
                    # A sheet with this name already exist
                    pass

                import time
                sheet.Cells(1,1).Value = "Exported from Scicraft on " + time.ctime()
                
                sheet.Cells(3,1).Value = data.get_name()
                sheet.Cells(5,1).Value = data.get_dim_names()[0]
                sheet.Cells(3,3).Value = data.get_dim_names()[1]

                x,y = data.shape
                exlrange = sheet.Range(sheet.Cells(5,3),sheet.Cells(x+4,y+2))
                exlrange.Value = data.asarray()

                exlrange = sheet.Range(sheet.Cells(5,2),sheet.Cells(x+4,2))
                exlrange.Value = [[x] for x in data.get_identifiers(data.get_dim_names()[0])]
                exlrange = sheet.Range(sheet.Cells(4,3),sheet.Cells(4,2+y))
                exlrange.Value = data.get_identifiers(data.get_dim_names()[1])                
                            
                excel.visible = 1
        except Exception:
            logger.log('warning','Putting data into Excel failed.')             
            
                           

""" Reads simple text files"""
class TextFileReader:    
    
    def insert_into(self, proj):
        data = self.run()
        if data:
            fun = "Textfiel Reader"
            if hasattr(self,"filename"):
                fun = os.path.basename(self.filename)
            proj.add_data(None,data,fun=fun)
        
    def run(self):
        WORKFLOWDIR = os.path.split(__file__)[0]
        
        self.founddata = None
        
        xml = gtk.glade.XML(os.path.join(WORKFLOWDIR,"textfilereader.glade"), 'dialog1')
    
        dialog = xml.get_widget('dialog1')
        self.filechooser = xml.get_widget("filechooserbutton1")            
        self.statuslabel = xml.get_widget("status_label")
        self.dataname_entry = xml.get_widget('dataset_name_entry')
        self.force_col_headers = xml.get_widget('force_col_headers')
        self.force_row_headers = xml.get_widget('force_row_headers')
        row_dim_entry = xml.get_widget('row_name_entry')
        col_dim_entry = xml.get_widget('col_name_entry')
        
        self.filechooser.connect("selection-changed", self._on_settings_changed)
        self.force_col_headers.connect("toggled", self._on_settings_changed)
        self.force_row_headers.connect("toggled", self._on_settings_changed)
        
        try:
            status = dialog.run()
        finally:
            dialog.destroy()
        try:    
            if self.founddata and status != gtk.RESPONSE_DELETE_EVENT:
                data, row_ids, col_ids = self.founddata        
                datasetname = self.dataname_entry.get_text()
                row_dim_name  = row_dim_entry.get_text()
                col_dim_name  = col_dim_entry.get_text()
                a = [dataset.Dataset(data,[(row_dim_name, row_ids),(col_dim_name,col_ids)],name=datasetname)]                                
        except Exception:
            logger.log("warning","Error in R when reading text. Function aborted")
        return a            

            
 
    def _set_status_successfull(self,dimensions,rows,cols):
        text = "\nDataset: Valid\n\nDimensions: %s\n\nRow Names:%s\nColumn Names: %s" %\
                    ( ", ".join(map(str,dimensions)), ", ".join(rows), ", ".join(cols))
        self.statuslabel.set_label(text)
        
    def _set_status_invalid(self, rows, cols):
        text = "\nDataset: Invalid Dimensions\nThis was probably be caused by row- or column-headers being interpreted wrongly"+\
                        "\n\nRow Names: %s\nColumn Names: %s" %\
                        (", ".join(rows), ", ".join(cols))
        self.statuslabel.set_label(text)
                                    
    def _set_status_failed(self,filename):
        text = "\nDataset: ERROR\n\nFailure parsing %s" % filename ( dimensions, ", ".join(rows), ", ".join(cols))
        self.statuslabel.set_label(text)
                   
    def _on_settings_changed(self,file_button):
      
        filename = self.filechooser.get_filename()
        self.filename = filename # store for use 
        
        force_col = self.force_col_headers.get_active()
        force_row = self.force_row_headers.get_active()
        
        data_list, colids, rowids = self._parse(filename, force_col_headers=force_col,
                                                    force_row_headers=force_row)
        
        if not data_list:
            self._set_status_failed(filename)
            return
        
        # converting
        colids = map(str,colids)
        rowids = map(str,rowids)        
        
        # update to new suggested dataset name
        self.dataname_entry.set_text(filename.split(os.sep)[-1].split(".")[0])
        
        # try converting dataset, and update status message
        try:            
            self._convert_matrix(data_list)        
            a = scipy.array(data_list)
            a = scipy.transpose(a)
            dataset.Dataset(a,[(None,rowids), (None,colids)])   # test
            self.founddata = a, rowids, colids
            self._set_status_successfull(a.shape,rowids, colids)
        except dataset.IdentifierMismatchError:
            self._set_status_invalid(rowids, colids)                   
        except ValueError:
            self._set_status_invalid(rowids, colids)


    def _convert_matrix(self, list):
        for x, l in enumerate(list):
            if isinstance(l[0],str):
                list[x] = map(float,list[x])                
                
    # parses dataset, does not change datatypes, nor validate result. Dataset returned must be transposed
    def _parse(self,filename,force_col_headers = False, force_row_headers = False): 
        if os.name=="nt":
            filename = filename.replace("\\","/")

        rpy.r('a = read.table(file("%s"))' % (filename))
        data_dict = rpy.r("a")
        colnames  = rpy.r("colnames(a)")
        if not isinstance(colnames,list):
            colnames = [colnames]    
                
        data_list = []
        
        # force row headers?
        if force_row_headers:
            rownames = data_dict[colnames.pop(0)]
        else:
            rownames = rpy.r("rownames(a)")
    
        

        # creating list of lists.. this array will need to be transposed, 
        #but postponing this until we can create an actual array object                    
        for key in colnames:
            data_list.append(data_dict[key])
        
        # force col_headers?
        if force_col_headers:
            colnames = [row.pop(0) for row in data_list]
            rownames.pop(0)
        
        # can't create dataset, as we've not validated matrix nor changed types
        return (data_list,colnames, rownames)


class RLoadFunction:

    def insert_into(self, proj):
        data = self.run()
        if data:
            fun = "R Reader"
            if hasattr(self,"filename"):
                fun = os.path.basename(self.filename)
            proj.add_data(None,data,fun=fun)
            
    def run(self):      
        import rpy
                
        filename = dialogs.get_open_filename("Select R file...")
        self.filename = filename                
        
        if filename:
            if sys.platform == "win32":
                # rpy wants filenames in this format
                filename = filename.replace("\\","/")             
            loaded = rpy.r("a <- as.matrix(load(\"%s\"))" % (filename,))
            
            if isinstance(loaded,Numeric.arraytype):
                loaded = Numeric.ravel(loaded)
                                                                         
            if isinstance(loaded,str):
                loaded = [loaded]
                            
            datasets = []
            for name in loaded:
                name = str(name)
                colnames = rpy.r("colnames(%s)" % (name,))
                rownames = rpy.r("rownames(%s)" % (name,))
                data     = rpy.r("as.matrix(%s)" % (name,))
                datasets.append(dataset.Dataset(data,((None,rownames),(None,colnames)),name=name))
            return datasets 

class RSaveFunction:
    def run(self, data):
        import rpy
        dataname = data.get_name()
        rpy.r.assign(dataname, data.asarray())
        filename = dialogs.get_save_filename("Select save file", data.get_name()+".RData",("R Dataset","RData"))
        if filename:
            if sys.platform == "win32":
                # rpy wants filenames in this format
                filename = filename.replace("\\","/") 
            rpy.r('save(%s,file="%s")' % (dataname, filename,))
            logger.log("notice", "Saved data to %r." % filename)        

class MatFileLoadFunction:
    """Loader for .mat-files."""
    def insert_into(self, proj):
        data = self.run()
        if data:
            fun = "Matlab Reader"
            if hasattr(self,"filename"):
                fun = os.path.basename(self.filename)
            proj.add_data(None,data,fun=fun)
            
    def run(self,filename=None):
        if not filename:
            filename = dialogs.get_open_filename("Select M file...")
        self.filename = filename
        ids = []           

        if filename:
            try:            
                loaded = matfile.read(filename)
            except IOError, e:
                try:
                    loaded = scipy.io.loadmat(filename)
                except IOError, e:
                    logger.log("error",e)
                    return

            data_set = []
            
            for name, data in loaded.items():
                if name=="__nargin__":
                    continue
                           
                if type(data)==str:
                    data_set.append(project.Notice(name,data))
                    
                else:
                
                    if data.typecode()=='1' or data.typecode()=='s':
                        data = Numeric.array(data,Numeric.Int)
                    if data.typecode()=='c':                                                               
                        # complex datatype!
                        ids.append((name, data))                                                                                                
                    else:
                        data_set.append(dataset.Dataset(data,name=name))                           
            
        if ids:
            dia = gtk.Dialog("Apply Row/Column Names",buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OK,gtk.RESPONSE_ACCEPT))
            vbox = gtk.VBox()
            dia.vbox.pack_start(vbox)
            stringframe = gtk.Frame("Found String-Vectors")
            vbox.pack_start(stringframe)
            table = gtk.Table()
            stringframe.add(table)
            
            
            lengths  = {}
            datas    = {}
            for x,(name, data) in enumerate(ids):
                vector = ["".join(map(str,line)).replace(chr(0),"") for line in data]
                table.attach(gtk.Label(name),0,1,x,x+1)
                table.attach(gtk.Label(len(vector)),1,2,x,x+1)
                entry = gtk.Entry()
                entry.set_text(" ".join(vector))
                entry.set_editable(False)
                table.attach(entry,2,3,x,x+1)
                
                # store for filling combobox'
                datas[name] = vector
                if lengths.has_key(len(vector)):
                    lengths[len(vector)].append(name)
                else:
                    lengths[len(vector)] = [name]                
    
            valueframe = gtk.Frame("Apply String-Vectors")
            vbox.pack_start(valueframe)
            table = gtk.Table()
            valueframe.add(table)            
            
            
            combos = []
            for x, data in enumerate(data_set):
                if isinstance(data,dataset.Dataset):
                    table.attach(gtk.Label(data.get_name()),0,1,x,x+1)
                    table.attach(gtk.Label(str(data.shape)),1,2,x,x+1)
                    for y,num in enumerate(data.shape):
                        combo = gtk.combo_box_new_text()
                         
                        if num in lengths.keys():
                            for name in lengths[num]:
                                combo.append_text(name)
                            combo.append_text("None")
                            combo.set_active(0)
                            combos.append((data,y,combo))
                        else:
                            combo.set_sensitive(False)
                                                                                
                        table.attach(combo,2+y,3+y,x,x+1)
                    
            dia.show_all()
            try:
                if dia.run()==gtk.RESPONSE_ACCEPT:            
                    dimnames = {}
                    for data, dimnum, combo in combos:
                        text = combo.get_active_text()
                        if text in datas:
                            data.change_identifiers(data.get_dim_names()[dimnum],datas[text])                          
                            # ensure equal identifiers --> equal dim-names
                            if text in dimnames:
                                data.change_dim_name(data.get_dim_names()[dimnum],dimnames[text])
                            else:
                                dimnames[text] = data.get_dim_names()[dimnum]                                 
            finally:
                dia.destroy() 
        return data_set                
              


class SaveMatlabFunction:
    """Loader for .mat-files."""

    def run(self,data,*more_datas):    

        filename = dialogs.get_save_filename("Select save file",data.get_name()+".mat",("Matlab","mat"))
        if filename:
            # if numpy is available, scipy.io.savemat will refuse to recognize anything else
            try:
                import numpy as numx
            except ImportError:
                import Numeric as numx

            arrays = {data.get_name() : numx.array(data.asarray())}                                      
            scipy.io.savemat(filename, arrays)
            logger.log("notice", "Saved data to %r." % filename)
        
class DatasetSaveFunction:
    """QND way to save data to file for later import to this program."""
    def run(self, data):
        filename = dialogs.get_save_filename("Select save file",data.get_name()+".pkl",("Python Objects","pkl"))
        if filename:
            cPickle.dump(data, open(filename, "wb"), protocol=2)
            logger.log("notice", "Saved data to %r." % filename)

class ExportTableFunction:
    def run(self, data):
        text = self.convert((data,))
        filename = dialogs.get_save_filename("Save tab-separated data...",
                                             data.get_name() + ".txt")
        if filename:
            open(filename, "w").write(text)
            logger.log("notice", "Saved data to %r." % filename)

    def convert(self, datalist, selection=None):
        if len(datalist) == 1:
            dim_1, dim_2 = datalist[0].get_dim_names()
            if len(datalist[0].get_identifiers(dim_1)) > len(datalist[0].get_identifiers(dim_2)):
                dim = dim_1
            else:
                dim = dim_2
        else:

            dims = set(datalist[0].get_dim_names())
            for d in list(dims):
                if len(datalist[0].get_identifiers(d)) == 1:
                    dims.remove(d)
                    
            for data in datalist[1:]:
                dims.intersection_update(data.get_dim_names())

            dim, = dims

        buf = StringIO()

        rownames = datalist[0].get_identifiers(dim)

        column_names = [dim]
        columns = []
        for data in datalist:
            names = data.get_dim_names()
            names.remove(dim)
            dim_name, = names
            if 1 in data.asarray().shape:
                column_names.append(data.get_name())
            elif len(datalist) == 1:
                    column_names.extend(data.get_identifiers(dim_name))
            else:
                if 1 in datalist[0].asarray().shape:
                    column_names.append(datalist[0].get_name())
                column_names.extend(["%s_%s" % (dim_name, ident) for ident in data.get_identifiers(dim_name)])

            if dim == data.get_dim_names()[1]:
                columns.extend(data.asarray().tolist())
            else:
                columns.extend(zip(*data.asarray().tolist()))
            
        print >> buf, "\t".join(column_names)

        print_all = not selection or dim not in selection or len(selection[dim]) == 0

        for rowname, row in zip(rownames, zip(*columns)):
            if print_all or rowname in selection[dim]:
                print >> buf, "\t".join([rowname] + map(str, row))

        return buf.getvalue()                                 
            
     
