#!/usr/bin/env python

import pygtk
from system import plots, logger
import os
from gtk.gtkgl.apputils import *
from math import sqrt, fabs

class MeshPlotBox(plots.Plot):
    def __init__(self, data1, target=500):
        plots.Plot.__init__(self, 'Mesh Plot')
        if sys.platform != 'win32':
            self.set_resize_mode(gtk.RESIZE_IMMEDIATE)
        self.set_reallocate_redraws(True)
        self.connect('destroy', lambda quit: gtk.main_quit())
        self.connect('parent_set', self.parentset)


        self.content = MeshPlot(data1, target, self)
        self.glarea = GLArea(self.content)
        self.glarea.show()
        
        self.init_toolbar()
        self.settings_window = None

        self.eventbox = gtk.EventBox()
        self.eventbox.add(self.glarea)
        self.eventbox.show()
        self.add(self.eventbox)
        
        self.new = True

        self.show()
        
    def init_toolbar(self):
        self._toolbar = gtk.Toolbar()
        #b = gtk.ToolButton(None, "Settings")
        #b.connect("clicked", self.toggle_settings)
        #self._toolbar.add(b)
        b = gtk.ToolButton(None, "Help")
        b.connect("clicked", self.show_help)
        self._toolbar.add(b)
        self._toolbar.show_all()
        
    def parentset(self, widget, event):
        self.content.mesh.printer.newParent()

    def radio_toggled(self, radio, title):
        if title == "Wireframe":
            self.content.fill = not radio.get_active()
        elif title == "Filled":
            self.content.fill = radio.get_active()

    def check_toggled(self, box, title):
        if title == "Threshold plane":
            self.content.mesh.show_thresh = box.get_active()
        elif title == "Axis labels":
            self.content.mesh.show_markers = box.get_active()
        
    def spinner_changed(self, spinner, title):
        if title == "Threshold value":
            self.content.mesh.thresh = spinner.get_value()
        elif title == "Y scale":
            self.content.mesh.yScale = spinner.get_value()
            
    def scale_changed(self, scale, title):
        if title == "Horizontal rotation":
            self.content.roty = scale.get_value()
        elif title == "Vertical rotation":
            self.content.rotx = scale.get_value()
        elif title == "Zoom":
            self.content.mesh.zoom = scale.get_value()
            
    def on_key_press(self, widget, event):
        if event.keyval==43: # + 
            self.content.mesh.yScale += 0.1
        elif event.keyval==45: # -
            self.content.mesh.yScale -= 0.1
        elif event.keyval==119: # w
            self.content.fill = not self.content.fill
        elif event.keyval==113: # q
            self.content.mesh.raise_thresh()
        elif event.keyval==97: # a
            self.content.mesh.lower_thresh()
        elif event.keyval==122: # z
            self.content.mesh.show_thresh = not self.content.mesh.show_thresh
        elif event.keyval==109: # m
            self.content.mesh.show_markers = not self.content.mesh.show_markers
        elif event.keyval==65513: #alt
            self.content.pan_active = True
        elif event.keyval==65507: #ctrl
            self.content.zoom_active = True
            
    def on_key_release(self, widget, event):
        if event.keyval==65513: #alt
            self.content.pan_active = False
        elif event.keyval==65507: #ctrl
            self.content.zoom_active = False
        
    def run(self):
        self.show()
        gtk.main()
        
    def __getitem__(self, key):
        return self._widget_tree.get_widget(key)

    def get_toolbar(self):
        return self._toolbar

    def toggle_settings(self, button):
        if self.settings_window:
            self.settings_window.destroy()
        else:
            DATADIR = os.path.dirname(sys.modules['system'].__file__)
            self.settings_window = gtk.Window()
            notebook = gtk.Notebook()
            self.settings_window.add(notebook)

            m = MeshPlotSettings(self)
            notebook.prepend_page(m, m.title)
            m = MeshPlotHelp(self)
            notebook.append_page(m, m.title)
            self.settings_window.connect('destroy', self.settings_closed)
            self.settings_window.show_all()

    def settings_closed(self, window):
        self.settings_window = None

    def show_help(self, button):
        help_window = gtk.Window()
        help_window.set_title("Help")
        content = gtk.Label("Help! I've fallen and can't get up!")
        help_window.add(content)
        help_window.show_all()
    
    def get_name(self):
        return "Mesh plot"
        
class MeshPlotSettings(gtk.VBox):

    def __init__(self, window_settings):
        gtk.VBox.__init__(self,"Mesh plot settings")
        self._window_settings = window_settings
        self.title = gtk.Label("Settings")
        
        settingsframe = gtk.Frame("Mesh plot settings")
        settingsframe.get_label_widget().set_markup("<b>Mesh settings</b>")
        self.pack_start(settingsframe, expand=False)

        self.settings_table = gtk.Table(columns=3, rows=4)
        self.settings_table.attach(gtk.Label("Wireframe"),0, 1, 1, 2)
        self.settings_table.attach(gtk.Label("Filled"), 0, 1, 2, 3)
        radionum = 2
        b1 = gtk.RadioButton(None)
        b2 = gtk.RadioButton(b1)
        b1.connect("toggled", self._window_settings.radio_toggled, "Wireframe")
        b2.connect("toggled", self._window_settings.radio_toggled, "Filled")
        fill = self._window_settings.content.fill
        if fill:
            b2.set_active(True)
        else:
            b1.set_active(True)
                
        self.settings_table.attach(b1, 1, 2, 1, 2)
        self.settings_table.attach(b2, 1, 2, 2, 3)

        val = self._window_settings.content.mesh.show_markers
        self.settings_table.attach(gtk.Label("Axis labels"),0,1,radionum+1,radionum+2)
        b = gtk.CheckButton()
        if val:
            b.set_active(True)
        b.connect("toggled", self._window_settings.check_toggled, "Axis labels")
        self.settings_table.attach(b,1,2, radionum+1, radionum+2)

        val = self._window_settings.content.mesh.show_thresh
        self.settings_table.attach(gtk.Label("Threshold plane "),0,1,radionum+2,radionum+3)
        b = gtk.CheckButton()
        if val:
            b.set_active(True)
        b.connect("toggled", self._window_settings.check_toggled, "Threshold plane")
        self.settings_table.attach(b,1,2, radionum+2, radionum+3)

        cur = self._window_settings.content.mesh.thresh
        mn = self._window_settings.content.mesh.mn
        mx = self._window_settings.content.mesh.mx
        b = gtk.SpinButton(gtk.Adjustment(cur, mn, mx, 0.1, 0, 0), 0.05, 4)
        b.connect("value-changed", self._window_settings.spinner_changed, "Threshold value")
        self.settings_table.attach(b, 2, 3, radionum + 2, radionum + 3)

        scale = self._window_settings.content.mesh.yScale
        b = gtk.SpinButton(gtk.Adjustment(scale, 0.0, 25.0, 0.1, 0, 0), 0.1, 2)
        b.connect("value-changed", self._window_settings.spinner_changed, "Y scale")
        self.settings_table.attach(gtk.Label("Value scale:"), 0, 1, radionum+3, radionum+4)
        self.settings_table.attach(b, 1, 2, radionum + 3, radionum + 4)
        settingsframe.add(self.settings_table)
        self.show_all()

        
class MeshPlotHelp(gtk.VBox):
    def __init__(self, window_settings):
        gtk.VBox.__init__(self,"Mesh plot settings")
        self._window_settings = window_settings
        self.title = gtk.Label("Help")
        
        helpframe = gtk.Frame("Mesh plot instructions")
        helpframe.get_label_widget().set_markup("<b>Plot instructions</b>")
        self.pack_start(helpframe, expand=False)

        self.help_table = gtk.Table(columns=2, rows=4)
        l = gtk.Label("Left mouse button - Rotate mesh\n"+
                      "Middle mouse button - Zoom\n"+
                      "Right mouse button - Pan camera\n\n"+
                      "Keyboard commands require a separate window\n"+
                      "W - Switch between wireframe and polygonal rendering\n"+
                      "+ - Increase value scaling\n- - Decrease value scaling\n"+
                      "Z - Toggle threshold plane on/off\nQ - Raise threshold plane\n"+
                      "A - Lower threshold plane\n"+
                      "M - Toggle axis labels")
        self.help_table.attach(l, 0, 1, 0, 1)
        helpframe.add(self.help_table)
        self.show_all()

class CameraSettings(gtk.VBox):
    def __init__(self, parent):
        gtk.VBox.__init__(self, "Camera settings")
        self._parent = parent
        self.title = gtk.Label("Camera")

        cameraframe = gtk.Frame("Camera settings")
        cameraframe.get_label_widget().set_markup("<b>Camera settings</b>")
        self.pack_start(cameraframe, expand=False)

        self.camera_table = gtk.Table(columns=4, rows=4)
        s = gtk.HScale(gtk.Adjustment(30, -180, 180, 10, 0, 0))
        s.set_draw_value(False)
        s.connect("value-changed", self._parent.scale_changed, "Horizontal rotation")
        self.camera_table.attach(s, 1, 3, 0, 1)

        s = gtk.VScale(gtk.Adjustment(30, -180, 180, 10, 0, 0))
        s.set_draw_value(False)
        s.connect("value-changed", self._parent.scale_changed, "Vertical rotation")
        self.camera_table.attach(s, 0, 1, 1, 3)

        s = gtk.VScale(gtk.Adjustment(30, 0, 180, 10, 0, 0))
        s.set_draw_value(False)
        s.connect("value-changed", self._parent.scale_changed, "Zoom")
        self.camera_table.attach(s, 1, 2, 1, 3)
        
        cameraframe.add(self.camera_table)
        self.show_all()
        
class Mesh:
    def __init__(self, data, target):
        ar = data.asarray()
        dims = data.get_dim_names()
        self.x_dim = dims[0]
        self.z_dim = dims[1]
        self.width=len(ar)
        self.height=len(ar[0])
        print str(self.width), str(self.height)
        
        # Limit mesh complexity to target width and height
        
        #self.target = 25
        #if self.width>self.target:
        #    self.skip_w = float(self.width)/float(self.target)
        #    self.width = self.target
        #else:
        #    self.skip_w = 1.0
        #if self.height>self.target:
        #    self.skip_h = float(self.height)/float(self.target)
        #    self.height = self.target
        #else:
        #    self.skip_h = 1.0
        self.skip_w = 1.0
        self.skip_h = 1.0
        
        self.indices = [[0,0,0] for i in range(self.width*self.height*2)]
        self.vertices = [[0,0,0] for i in range(self.width*self.height)]
        self.colors = [[0,0,0] for i in range(self.width*self.height)]
        self.normals = [[0,0,0] for i in range(self.width*self.height)]
        
        c = 0
        self.mx = ar[0][0]
        self.mn = ar[0][0]

        self.sch = 20.0/float(self.height)
        self.scw = 20.0/float(self.width)
        
        self.midx = self.width/2 * self.scw
        self.midz = self.height/2 * self.sch
        for j in range(self.height):
            for i in range(self.width):
                self.vertices[c][0] = i*self.scw
                self.vertices[c][1] = ar[int(i*self.skip_w)][int(j*self.skip_h)]
                self.vertices[c][2] = j*self.sch
                if ar[int(i*self.skip_w)][int(j*self.skip_h)]>self.mx:
                    self.mx = ar[int(i*self.skip_w)][int(j*self.skip_h)]
                if ar[int(i*self.skip_w)][int(j*self.skip_h)]<self.mn:
                    self.mn = ar[int(i*self.skip_w)][int(j*self.skip_h)]
                c += 1
                
        c = 0
        for j in range(1, self.height-1):
            for i in range(1, self.width-1):
                c = i+j*self.width
                l = c-1
                r = c+1
                u = i+(j-1)*self.width
                d = i+(j+1)*self.width
                hor = [self.vertices[r][0]-self.vertices[l][0], self.vertices[r][1]-self.vertices[l][1], self.vertices[r][2]-self.vertices[l][2]]
                ver = [self.vertices[d][0]-self.vertices[u][0], self.vertices[d][1]-self.vertices[u][1], self.vertices[d][2]-self.vertices[u][2]]
                nor = [hor[1]*ver[2]-hor[2]*ver[1], hor[2]*ver[0]-hor[0]*ver[2], hor[0]*ver[1]-hor[1]*ver[0]]
                l = sqrt(nor[0]**2 + nor[1]**2 + nor[2]**2)
                nor[0] /= -l
                nor[1] /= -l
                nor[2] /= -l
                self.normals[c] = nor
                
        for j in range(self.height):
            c1 = int(j*self.width)
            c2 = int(self.width-1 + j*self.width)
            self.normals[c1] = self.normals[c1+1]
            self.normals[c2] = self.normals[c2-1]
        for i in range(self.width):
            c1 = i
            c2 = int(i+(self.height-1)*self.width)
            self.normals[c1] = self.normals[c1+int(self.width)]
            self.normals[c2] = self.normals[c2-int(self.width)]
                
        c = 0
        rng = (self.mx-self.mn)*2
        for j in range(self.height):
            for i in range(self.width):
                self.colors[c][0] = 0.7
                self.colors[c][1] = 0.7
                self.colors[c][2] = 0.7
                c += 1
        c = 0      
        for j in range(self.height-1):
            for i in range(self.width-1):
                self.indices[c][0] = i+j*self.width
                self.indices[c][1] = i+1+j*self.width
                self.indices[c][2] = i+1+(j+1)*self.width
                
                self.indices[c+1][0] = i+1+(j+1)*self.width
                self.indices[c+1][1] = i+(j+1)*self.width
                self.indices[c+1][2] = i+j*self.width
                c += 2

        self.curInd = self.indices[:]
        self.curWidth = self.width
        self.curHeight = self.height
       
        self.axisverts = [[0,0,0] for i in range(4)]
        self.axiscols = [[0,0,0] for i in range(4)]
        self.axisind = [[0,0] for i in range(3)]
        self.axisverts[0] = [0,0,0]
        self.axisverts[1] = [0.25, 0, 0]
        self.axisverts[2] = [0, 0.25, 0]
        self.axisverts[3] = [0, 0, 0.25]
        self.axiscols[0] = [1, 1, 1]
        self.axiscols[1] = [1, 0, 0]
        self.axiscols[2] = [0, 1, 0]
        self.axiscols[3] = [0, 0, 1]
        self.axisind[0] = [0, 1]
        self.axisind[1] = [0, 2]
        self.axisind[2] = [0, 3]
        
        self.bbverts = [[0,0,0] for i in range(8)]
        self.bbcols = [[1,1,1] for i in range(8)]
        self.bbind = [[0,0] for i in range(12)] 
        self.bbverts[0]=[0, self.mn, 0]    
        self.bbverts[1]=[(self.width-1)*self.scw, self.mn, 0]    
        self.bbverts[2]=[(self.width-1)*self.scw, self.mn, (self.height-1)*self.sch]    
        self.bbverts[3]=[0, self.mn, (self.height-1)*self.sch]    
        self.bbverts[4]=[0, self.mx, 0]    
        self.bbverts[5]=[(self.width-1)*self.scw, self.mx, 0]    
        self.bbverts[6]=[(self.width-1)*self.scw, self.mx, (self.height-1)*self.sch]    
        self.bbverts[7]=[0, self.mx, (self.height-1)*self.sch]   
        self.bbind[0] = [0, 1]   
        self.bbind[1] = [1, 2]   
        self.bbind[2] = [2, 3]   
        self.bbind[3] = [3, 0]   
        self.bbind[4] = [4, 5]   
        self.bbind[5] = [5, 6]   
        self.bbind[6] = [6, 7]   
        self.bbind[7] = [7, 4]   
        self.bbind[8] = [0, 4]   
        self.bbind[9] = [1, 5]   
        self.bbind[10] = [2, 6]  
        self.bbind[11] = [3, 7]
 
        self.printer = TextPrinter(os.sep.join(["system","plot3d","font.tga"]), (16,16))
        self.yScale = 1.0
        self.zoom = 30
        self.thresh = self.mn
        self.show_thresh = False
        self.show_markers = False
        self.rendering = GL_TRIANGLES
        self.WIREFRAME = 0
        self.SOLID = 1
        self.SHADED = 2
        self.render_method = self.SOLID
        self.linecolor = 1
        self.color1 = gtk.gdk.Color(0,0,0)
        self.color2 = gtk.gdk.Color(0,0,0)
        self.listChanged = True
        self.flatList = glGenLists(1)
        self.wireList = glGenLists(1)
        self.shadedList = glGenLists(1)
        print "Wired: ", self.wireList, "Flat: ", self.flatList, "Shaded: ", self.shadedList
        print("Mesh of size %sx%s generated."%(self.width, self.height))
                
    def drawAxes(self, rotx, roty):
        glPushMatrix()
        glLoadIdentity()
        glDisable(GL_DEPTH_TEST)
        glTranslatef(-1,-0.75,-1.75)
        glRotatef(rotx, 1, 0, 0)
        glRotatef(roty, 0, 1, 0)
        glVertexPointerf(self.axisverts)
    	glColorPointerf(self.axiscols)
    	glDrawElementsui(GL_LINES, self.axisind)
    	self.printer.draw_string("x", 0.25, 0, 0, (self.linecolor,self.linecolor,self.linecolor))
    	self.printer.draw_string("y", 0, 0.25, 0, (self.linecolor,self.linecolor,self.linecolor))
    	self.printer.draw_string("z", 0, 0, 0.25, (self.linecolor,self.linecolor,self.linecolor))
        glEnable(GL_DEPTH_TEST)
        glPopMatrix()
        
    def drawBB(self):
        # Bounding box
        glVertexPointerf(self.bbverts)
    	glColorPointerf(self.bbcols)
    	glDrawElementsui(GL_LINES, self.bbind)
    	
    	
        glBegin(GL_LINES)
        glVertex3f(0, self.thresh, 0)
        glVertex3f(-0.25, self.thresh, -0.25)
        glEnd()
        size = 0.8
        if self.show_thresh:
            self.printer.draw_string(str(round(self.thresh, 1)), -0.25, self.thresh, -0.25, (self.linecolor,self.linecolor,self.linecolor))
    	    if self.thresh < self.mn:
    	        self.thresh = self.mn
    	    elif self.thresh > self.mx:
    	        self.thresh = self.mx
    	    glColor3f(0.5, 0.5, 0.5)
    	    glBegin(GL_QUADS)
    	    glVertex3f(0, self.thresh, 0)
    	    glVertex3f((self.width-1)*self.scw, self.thresh, 0)
    	    glVertex3f((self.width-1)*self.scw, self.thresh, (self.height-1)*self.sch)
    	    glVertex3f(0, self.thresh, (self.height-1)*self.sch)
    	    glEnd()
    	
    	if self.show_markers:
           
    	    #glDisable(GL_DEPTH_TEST)
            steps = int(60.0/self.zoom) + 1
            rng = (self.mx-self.mn)
        	# Value markers y
            ySteps = int(steps*self.yScale/1.25)
            if ySteps==0:
                ySteps = 1
            incr = (rng/(ySteps))
            for i in range(int(steps*self.yScale/1.25)+1):
                val = round(i*incr+self.mn, 1)
                glBegin(GL_LINES)
                glVertex3f(0,i*incr + self.mn,0)
                glVertex3f(-0.25, i*incr + self.mn, -0.25)
                glEnd()
                self.printer.draw_string(str(val), -0.25, val, -0.25, (self.linecolor,self.linecolor,self.linecolor))
        
            # Value markers x
            self.printer.draw_string_3d(self.x_dim, (20, self.mn-1.0/self.yScale, 0),(-1,0,0), (self.linecolor,self.linecolor,self.linecolor))
            incr = float((self.width-1)*self.scw) / (steps+2)
            for i in range(1,steps+3):
                val = i * incr
                glBegin(GL_LINES)
                glVertex3f(val, self.mn, 0)
                glVertex3f(val, self.mn, -0.25)
                glEnd()
                self.printer.draw_string(str(int(round((val/self.scw+1) * self.skip_w,0))), val, self.mn, -0.25, (self.linecolor,self.linecolor,self.linecolor))
            
            # Value markers z
            self.printer.draw_string_3d(self.z_dim, (0, self.mn-1.0/self.yScale, 20 - len(self.z_dim)),(0,0,1), (self.linecolor,self.linecolor,self.linecolor))
    	    incr = float((self.height-1)*self.sch) / (steps+2)
            for i in range(1,steps+3):
                val = i * incr 
                glBegin(GL_LINES)
                glVertex3f(0, self.mn, val)
                glVertex3f(-0.25, self.mn, val)
                glEnd()
                self.printer.draw_string(str(int(round((val/self.sch+1) * self.skip_h,0))), -0.25, self.mn, val, (self.linecolor,self.linecolor,self.linecolor))
            #glEnable(GL_DEPTH_TEST)
        
    def render(self, fill):
        if self.listChanged:
        
            self.flatList = glGenLists(1)
            self.wireList = glGenLists(1)
            self.shadedList = glGenLists(1)
            print "Wired: ", self.wireList, "Flat: ", self.flatList, "Shaded: ", self.shadedList
            glNewList(self.wireList, GL_COMPILE)
            glColor3f(self.linecolor, self.linecolor, self.linecolor)
            glDisable(GL_LIGHTING)
            glDisable(GL_NORMAL_ARRAY)
            glDisable(GL_COLOR_ARRAY)
            glVertexPointerf(self.vertices)
            glDrawElementsui(GL_LINES, self.curInd)
            glEndList()
            
            glNewList(self.flatList, GL_COMPILE)
            glDisable(GL_LIGHT0)
            glDisable(GL_LIGHTING)
            glDisable(GL_NORMAL_ARRAY)
            glEnable(GL_COLOR_ARRAY)
            glVertexPointerf(self.vertices)
            glColorPointerf(self.colors)
            glDrawElementsui(GL_TRIANGLES, self.curInd)
            glEndList()
            
            glNewList(self.shadedList, GL_COMPILE)
            glEnable(GL_COLOR_ARRAY)
            glEnable(GL_NORMAL_ARRAY)
            glEnable(GL_LIGHTING)
            glEnable(GL_LIGHT0)
            glEnable(GL_COLOR_MATERIAL)
            glVertexPointerf(self.vertices)
            glColorPointerf(self.colors)
            glNormalPointerf(self.normals)
            glDrawElementsui(GL_TRIANGLES, self.curInd)
            glEndList()
            self.listChanged = False
        elif self.render_method == self.WIREFRAME:
            glCallList(self.wireList)
        elif self.render_method == self.SOLID:
            glCallList(self.flatList)
        elif self.render_method == self.SHADED:
            glCallList(self.shadedList)

    def raise_thresh(self):
        self.thresh += (self.mx-self.mn)/20.0

    def lower_thresh(self):
        self.thresh -= (self.mx-self.mn)/20.0

    def set_thresh(self, value):
        self.thresh = value

    def get_thresh(self):
        return self.thresh
    
    def toggle_thresh(self, value):
        self.show_thresh = value
        
    def set_rendering(self, title):
        if title=="Wireframe":
            self.rendering = GL_LINES
            self.render_method = self.WIREFRAME
        elif title=="Solid":
            self.rendering = GL_TRIANGLES
            self.render_method = self.SOLID
        elif title=="Shaded":
            self.rendering = GL_TRIANGLES
            self.render_method = self.SHADED

            
    def set_coloring(self, title, custom=None):
        self.listChanged = True
        if title=="Black <> White":
            rng = (self.mx-self.mn)
            c = 0
            for j in range(self.height):
                for i in range(self.width):
                    self.colors[c][0] = self.vertices[c][1]/rng
                    self.colors[c][1] = self.vertices[c][1]/rng
                    self.colors[c][2] = self.vertices[c][1]/rng
                    c += 1
        elif title=="Red <> Blue":
            rng = (self.mx-self.mn)
            c = 0
            for j in range(self.height):
                for i in range(self.width):
                    self.colors[c][0] = 1.0-self.vertices[c][1]/rng
                    self.colors[c][1] = 0
                    self.colors[c][2] = self.vertices[c][1]/rng
                    c += 1
        elif title=="Gray":
            c = 0
            for j in range(self.height):
                for i in range(self.width):
                    self.colors[c][0] = 0.6
                    self.colors[c][1] = 0.6
                    self.colors[c][2] = 0.6
                    c += 1
        elif title=="Custom":
            rng = (self.mx-self.mn)
            self.color1 = custom[0].get_color()
            self.color2 = custom[1].get_color()
            discrete = custom[2].get_active()
            count = custom[3].get_value()-1
            color = [[0,0,0],[0,0,0]]
            color[0][0] = float(self.color1.red) / 65535
            color[0][1] = float(self.color1.green) / 65535
            color[0][2] = float(self.color1.blue) / 65535
            color[1][0] = float(self.color2.red) / 65535
            color[1][1] = float(self.color2.green) / 65535
            color[1][2] = float(self.color2.blue) / 65535
            c = 0
            for j in range(self.height):
                for i in range(self.width):
                    val = self.vertices[c][1]/rng
                    if discrete:
                        val = float(int(val * (count) + 0.5))/float(count)
                    self.colors[c][0] = val*color[0][0] + (1-val)*color[1][0]
                    self.colors[c][1] = val*color[0][1] + (1-val)*color[1][1]
                    self.colors[c][2] = val*color[0][2] + (1-val)*color[1][2]
                    c += 1
        elif title=="Toggle print colors":
            self.linecolor = (self.linecolor + 1) % 2
            for i in range(8):
                val = (self.bbcols[i][0] + 1) % 2
                self.bbcols[i][0] = val
                self.bbcols[i][1] = val
                self.bbcols[i][2] = val

    def geometry_detail(self, widget, width, height):
        self.listChanged = True
        x = int(width.get_value())
        z = int(height.get_value())
        print "Adjusting detail level to ", str(x), str(z), str(self.width), str(self.height)
        self.skip_w = self.width/x
        self.skip_h = self.height/z
        print "Skipping: ", str(self.skip_w), str(self.skip_h)
        self.curInd = [[0,0,0] for i in range(x*z*2)]
        self.curWidth = x
        self.curHeight = z
        c = 0

        for i in range(x-1):
            for j in range(z-1):
                if i == x-2:
                    iplus = self.width - 1
                else:
                    iplus = (i+1)*self.skip_w
                if j == z-2:
                    zplus = self.height - 1
                else:
                    zplus = (j+1)*self.skip_h
                self.curInd[c][0] = int((i)*self.skip_w) + int((j)*self.skip_h)*self.width
                self.curInd[c][1] = int(iplus) + int((j)*self.skip_h)*self.width
                self.curInd[c][2] = int(iplus) + int(zplus)*self.width
                
                self.curInd[c+1][0] = int(iplus) + int(zplus)*self.width
                self.curInd[c+1][1] = int((i)*self.skip_w) + int(zplus)*self.width
                self.curInd[c+1][2] = int((i)*self.skip_w) + int((j)*self.skip_h)*self.width
                c += 2                

class MeshPlot(GLScene,
             GLSceneButton,
             GLSceneButtonMotion,
             GLSceneIdle):

    def __init__(self, data, target, plotbox):
        GLScene.__init__(self,
                         gtk.gdkgl.MODE_RGB   |
                         gtk.gdkgl.MODE_DEPTH |
                         gtk.gdkgl.MODE_DOUBLE)

        self.zoom_active = False
        self.pan_active = False
        self.plotbox = plotbox
        self.mesh = Mesh(data, target)
        self.rotx = 30
        self.roty = 30
        self.pan = self.mesh.midx
        self.elevate = 0
        self.aspect = 1.33
        self.fill = True
        self.beginx = 0
        self.beginy = 0
        self.mesh.set_rendering("Solid")
        self.mesh.set_coloring("Black <> White")
        self.clearcolor = 0
        self.perspective = True

        
        self.context_menu = gtk.Menu() # Context menu
        item = gtk.MenuItem("Camera")
        submenu = gtk.Menu()
        item_child = gtk.MenuItem("Perspective projection")
        item_child.connect("activate", self.context_selection, "camera", "Perspective")
        submenu.attach(item_child, 0,1,0,1)
        item_child = gtk.MenuItem("Parallell projection")
        item_child.connect("activate", self.context_selection, "camera", "Parallell")
        submenu.attach(item_child, 0,1,1,2)
        item.set_submenu(submenu)
        self.context_menu.attach(item, 0,1,0,1)
        
        item = gtk.MenuItem("Colors")
        submenu = gtk.Menu()
        item_child = gtk.MenuItem("Black <> White")
        item_child.connect("activate", self.context_selection, "color", "Black <> White")
        submenu.attach(item_child, 0,1,0,1)
        item_child = gtk.MenuItem("Red <> Blue")
        item_child.connect("activate", self.context_selection, "color", "Red <> Blue")
        submenu.attach(item_child, 0,1,1,2)
        item_child = gtk.MenuItem("Gray")
        item_child.connect("activate", self.context_selection, "color", "Gray")
        submenu.attach(item_child, 0,1,2,3)
        item_child = gtk.MenuItem("Custom")
        item_child.connect("activate", self.context_selection, "color", "Custom")
        submenu.attach(item_child, 0,1,3,4)
        item_child = gtk.MenuItem("Toggle print colors")
        item_child.connect("activate", self.context_selection, "color", "Toggle print colors")
        submenu.attach(item_child, 0,1,4,5)
        item.set_submenu(submenu)
        self.context_menu.attach(item, 0,1,1,2)

        item = gtk.MenuItem("Render Method")
        submenu = gtk.Menu()
        item_child = gtk.MenuItem("Wireframe")
        item_child.connect("activate", self.context_selection, "render", "Wireframe")
        submenu.attach(item_child, 0,1,0,1)
        item_child = gtk.MenuItem("Solid")
        item_child.connect("activate", self.context_selection, "render", "Solid")
        submenu.attach(item_child, 0,1,1,2)
        item_child = gtk.MenuItem("Shaded")
        item_child.connect("activate", self.context_selection, "render", "Shaded")
        submenu.attach(item_child, 0,1,2,3)
        item.set_submenu(submenu)
        self.context_menu.attach(item, 0,1,2,3)

        item = gtk.MenuItem("Geometry detail")
        item.connect("activate", self.context_selection, "toggle", "Geometry detail")
        self.context_menu.attach(item, 0,1,3,4)
        item = gtk.MenuItem("Threshold plane")
        item.connect("activate", self.context_selection, "toggle", "Threshold plane")
        self.context_menu.attach(item, 0,1,4,5)
        item = gtk.MenuItem("Toggle axis labels")
        item.connect("activate", self.context_selection, "toggle", "Toggle axis labels")
        self.context_menu.attach(item, 0,1,5,6)
  
    def context_selection(self, widget, action, title):
        def closedialog(w, d):
            d.destroy()
        if action=="render":
            self.mesh.set_rendering(title)
        elif action=="color":
            if title=="Custom":
                def set_custom_color(widget, title, values):
                    self.mesh.set_coloring(title, values)
                def toggle_discrete(toggle, spinner):
                    if toggle.get_active():
                        print "True"
                        spinner.set_editable(True)
                    else:
                        print "False"
                        spinner.set_editable(False)
                        
                dialog = gtk.Dialog(title="Custom colors")
                tlabel = gtk.Label("Select custom color mapping")
                table = gtk.Table(3,4)
                label1 = gtk.Label("High color")
                label2 = gtk.Label("Low color")
                label3 = gtk.Label("Discrete steps")
                discen = gtk.CheckButton("_Enable", True)
                discadjust = gtk.Adjustment(4, 2, 20, 1)
                disccount = gtk.SpinButton(discadjust, 1)
                discen.connect('toggled', toggle_discrete, disccount)
                applyButton = gtk.Button("Apply")
                closeButton = gtk.Button("Close")
                col1 = gtk.ColorButton()
                col1.set_color(self.mesh.color1)
                col2 = gtk.ColorButton()
                col2.set_color(self.mesh.color2)
                values = [col1, col2, discen, disccount]
                table.attach(label1, 0,1,0,1)
                table.attach(col1, 1,3,0,1)
                table.attach(label2, 0,1,1,2)
                table.attach(col2, 1,3,1,2)
                table.attach(label3, 0,1,2,3)
                table.attach(discen, 1,2,2,3)
                table.attach(disccount, 2,3,2,3)
                table.attach(applyButton, 1,2,3,4)
                table.attach(closeButton, 2,3,3,4)
                dialog.vbox.add(tlabel)
                dialog.vbox.add(table)
                applyButton.connect('clicked', set_custom_color, title, values)
                closeButton.connect('clicked', closedialog, dialog)
                dialog.show_all()
            elif title=="Toggle print colors":
                self.clearcolor = (self.clearcolor + 1) % 2
                self.mesh.set_coloring(title)
            else:
                self.mesh.set_coloring(title)
        elif action=="toggle":
            if title=="Toggle axis labels":
                self.mesh.show_markers = not self.mesh.show_markers
            elif title=="Threshold plane":
                def toggle_thresh(toggle):
                    self.mesh.toggle_thresh(toggle.get_active())
                def set_thresh(spinner):
                    self.mesh.set_thresh(spinner.get_value())
                    
                dialog = gtk.Dialog(title="Threshold plane")
                table = gtk.Table(2,2)
                label1 = gtk.Label("Toggle threshold plane")
                label2 = gtk.Label("Set threshold value")
                togglebutton = gtk.CheckButton()
                togglebutton.set_active(self.mesh.show_thresh)
                togglebutton.connect('toggled', toggle_thresh)
                spinadjust = gtk.Adjustment(self.mesh.get_thresh(), self.mesh.mn, self.mesh.mx, 0.1)
                spinner = gtk.SpinButton(spinadjust, 0.01, 2)
                spinner.connect('changed', set_thresh)
                closebutton = gtk.Button("Close")
                closebutton.connect('clicked', closedialog, dialog)
                table.attach(label1, 0,1,0,1)
                table.attach(label2, 0,1,1,2)
                table.attach(togglebutton, 1,2,0,1)
                table.attach(spinner, 1,2,1,2)
                dialog.vbox.add(table)
                dialog.vbox.add(closebutton)
                dialog.show_all()
                
                
            elif title=="Geometry detail":
                dialog = gtk.Dialog(title="Mesh detail")
                tlabel = gtk.Label("Select number of points to render")
                table = gtk.Table(2, 3)
                wlabel = gtk.Label("Width points")
                hlabel = gtk.Label("Height points")
                wadjust = gtk.Adjustment(self.mesh.curWidth, 20, self.mesh.width, 1.0, 0, 0)
                hadjust = gtk.Adjustment(self.mesh.curHeight, 20, self.mesh.height, 1.0, 0, 0)
                wspin = gtk.SpinButton(wadjust, 1, 0)
                hspin = gtk.SpinButton(hadjust, 1, 0)
                applyButton = gtk.Button("Apply")
                applyButton.connect('clicked', self.mesh.geometry_detail, wspin, hspin)
                closeButton = gtk.Button("Close")
                closeButton.connect('clicked', closedialog, dialog)
                table.attach(wlabel, 0, 1, 0, 1)
                table.attach(hlabel, 0, 1, 1, 2)
                table.attach(wspin, 1, 2, 0, 1)
                table.attach(hspin, 1, 2, 1, 2)
                table.attach(applyButton, 0, 1, 2, 3)
                table.attach(closeButton, 1, 2, 2, 3)
                dialog.vbox.add(tlabel)
                dialog.vbox.add(table)
                dialog.show_all()
        elif action=="camera":
            self.set_camera(title)
            
    def init(self):
        glClearColor(0,0,0,1)
        glDisable(GL_CULL_FACE)
        glEnable(GL_DEPTH_TEST)
        glEnable(GL_VERTEX_ARRAY)
        glViewport(0, 0, 100, 100)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glMatrixMode(GL_MODELVIEW)
    
    def reshape(self, width, height):
        self.aspect = float(width)/float(height)
        glViewport(0,0, width, height)
        self.mesh.listChanged = True
        #self._size = width, height
    
    def display(self, width, height):
        glViewport(0,0, width, height)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        if self.perspective:
            gluPerspective(64.0, self.aspect, 0.1, 100)
        else:
            glOrtho(-12.0, 12.0, -12.0, 12.0, 0.1, 100)
        glMatrixMode(GL_MODELVIEW)
        glClearColor(self.clearcolor, self.clearcolor, self.clearcolor, 1)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glLoadIdentity()
        pos = [self.mesh.midx, 50, self.mesh.midz, 1]
        glTranslatef(-self.pan, self.elevate, -self.mesh.zoom)
        glTranslatef(self.mesh.midx, 0, self.mesh.midz)
        glRotatef(self.rotx, 1, 0, 0)
        glRotatef(self.roty, 0, 1, 0)
        glTranslatef(-self.mesh.midx, 0, -self.mesh.midz)
        glScalef(1.0, self.mesh.yScale, 1.0)
        glLightfv(GL_LIGHT0, GL_POSITION, pos)
        self.mesh.render(self.fill)
        self.mesh.drawBB()
        self.mesh.drawAxes(self.rotx, self.roty)
    	#glDisable(GL_DEPTH_TEST);
    	#glDrawElementsui(GL_LINE_STRIP, self.mesh.indices)
    
    def button_press(self, width, height, event):
        self.beginx = event.x
        self.beginy = event.y
        self.plotbox.get_parent().focus()
                           
    def button_release(self, width, height, event):
        if event.state & gtk.gdk.BUTTON3_MASK:
            difx = self.beginx - event.x
            dify = self.beginy - event.y
            if True:           
                self.context_menu.popup(None, None, None, 0, 0)
                self.context_menu.show_all()
    
    def button_motion(self, width, height, event):
        if event.state & gtk.gdk.BUTTON1_MASK:
            if self.pan_active:
                self.pan = self.pan + ((event.x-self.beginx))/2
                self.elevate = self.elevate + ((event.y-self.beginy))/2
            elif self.zoom_active:
                self.mesh.zoom = self.mesh.zoom + ((event.y-self.beginy))/2
            else:
                self.rotx = self.rotx + ((event.y-self.beginy)/width)*180
                self.roty = self.roty + ((event.x-self.beginx)/height)*180
        elif event.state & gtk.gdk.BUTTON2_MASK:
            self.mesh.zoom = self.mesh.zoom + ((event.y-self.beginy))/2
        self.beginx = event.x
        self.beginy = event.y
        self.update()
        self.invalidate()
        
    def idle (self, width, height):
        # Invalidate whole window.
        try:
            self.invalidate()
            self.display(width, height)
        except:
            pass
    
    
    def set_camera(self, title):
        if title=="Perspective":
            self.perspective = True
        elif title=="Parallell":
            self.perspective = False
     
        
        
                
class TextPrinter:
    LEFT, RIGHT = range(2)
    
    def __init__(self, imagepath, char_size=None, color=(1,1,1)):
        """Use square image as texture-base.
        
        Width/height of image must be power-of-2. If char_size is
        specified, characters will be scaled to (width, height) in
        pixels when printed.
        """
        self._imagepath = imagepath
        self._glid = None
        self._size = None
        self._color = color
        self.p = gtk.gdk.pixbuf_new_from_file(self._imagepath)
        if char_size:
            self._char_width, self._char_height = char_size
        else:
            # image is loaded twice (here and init_texture) instead of
            # keeping it buffered
            #image = gtk.gdk.pixbuf_new_from_file(imagepath)
            #self._char_width = image.get_width() / 16
            #self._char_height = image.get_height() / 16
            self._char_width = 16
            self._char_height = 16
        
        

    def newParent(self):
        self._glid = None
        
    def _init_texture(self):
        self._glid = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, self._glid)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        # allow texture to be mask and color to be set by glColor*
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)
        if self.p.get_n_channels() == 3:
            mode = GL_RGB
        else:
            mode = GL_RGBA
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, self.p.get_width(), self.p.get_height(), 0,
                     mode, GL_UNSIGNED_BYTE, self.p.get_pixels())

    def update_size(self, screen_size):
        """Needed for correct 2d-pixel translations in draw_string_2d().

        screen_size is a (width, height) tuple.
        """
        self._size = screen_size

    def _draw_letter_quad(self, obj_coords, letter):
        letter_value = ord(letter)
        # 16 letters per row in the texture
        letter_row = letter_value / 16
        letter_col = letter_value % 16

        # map to [0, 1]
        s1 = letter_col / 16.0
        s2 = s1 + (1/16.0)
        t1 = letter_row / 16.0
        t2 = t1 + (1/16.0)
        tex_coords = (s1, t2), (s2, t2), (s2, t1), (s1, t1)
        glBegin(GL_QUADS)
        for obj_coord, tex_coord in zip(obj_coords, tex_coords):
            glTexCoord(*tex_coord)
            glVertex(*obj_coord)
        glEnd()
    
    def draw_string(self, string, x, y, z, color, align=LEFT):
        if self._glid == None:
            self._init_texture()
        glPushMatrix()
        glColor3f(*color)
        length = len(string)
        middle = length/2
        
        #sglPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        glEnable(GL_TEXTURE_2D)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        
        glBindTexture(GL_TEXTURE_2D, self._glid)
        
        model_matrix = glGetDoublev(GL_MODELVIEW_MATRIX)
        projection_matrix = glGetDoublev(GL_PROJECTION_MATRIX)
        viewport = glGetIntegerv(GL_VIEWPORT)

        # get width/height in pixels of each letter
        width, height = self._char_width, self._char_height

        winx, winy, winz = gluProject(x, y, z, model_matrix, projection_matrix, viewport)
        winx, winy = int(winx), int(winy)

        if align == TextPrinter.RIGHT:
            winx -= width*len(string)
        
        for n, letter in enumerate(string):
            indent = width*n - middle*width
            # clockwise square
            pos_1 = gluUnProject(winx+indent, winy, winz, model_matrix, projection_matrix, viewport)
            pos_2 = gluUnProject(winx+indent+width, winy, winz, model_matrix, projection_matrix, viewport)
            pos_3 = gluUnProject(winx+indent+width, winy+height, winz, model_matrix, projection_matrix, viewport)
            pos_4 = gluUnProject(winx+indent, winy+height, winz, model_matrix, projection_matrix, viewport)
            self._draw_letter_quad([pos_1, pos_2, pos_3, pos_4], letter)
        glDisable(GL_BLEND)
        glDisable(GL_TEXTURE_2D)
        glPopMatrix()
        
    def draw_string_3d(self, string, start, direction, color, align=LEFT):
        if self._glid == None:
            self._init_texture()
        glPushMatrix()
        glColor3f(*color)
        length = len(string)
        middle = length/2
        startx, starty, startz = start[0], start[1], start[2]
        dirx, diry, dirz = direction[0], direction[1], direction[2]
        incrx = dirx
        incry = diry
        incrz = dirz
        #sglPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        glEnable(GL_TEXTURE_2D)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        
        glBindTexture(GL_TEXTURE_2D, self._glid)
        
        # get width/height in pixels of each letter
        width, height = self._char_width, self._char_height

        height = 1
        width = 1
        for n, letter in enumerate(string):
            indent = width*n - middle*width
            # clockwise square
            #pos_1 = gluUnProject(winx+indent, winy, winz, model_matrix, projection_matrix, viewport)
            #pos_2 = gluUnProject(winx+indent+width, winy, winz, model_matrix, projection_matrix, viewport)
            #pos_3 = gluUnProject(winx+indent+width, winy+height, winz, model_matrix, projection_matrix, viewport)
            #pos_4 = gluUnProject(winx+indent, winy+height, winz, model_matrix, projection_matrix, viewport)
            pos_1 = (startx + n*incrx, starty + n*incry, startz + n*incrz)
            pos_2 = (startx + (n+1)*incrx, starty + (n+1)*incry, startz + (n+1)*incrz)
            pos_3 = (startx + (n+1)*incrx, starty + (n+1)*incry + height, startz + (n+1)*incrz)
            pos_4 = (startx + n*incrx, starty + n*incry + height, startz + n*incrz)
            self._draw_letter_quad([pos_1, pos_2, pos_3, pos_4], letter)
        glDisable(GL_BLEND)
        glDisable(GL_TEXTURE_2D)
        glPopMatrix()
      
