import gtk
import sys
import os
import inspect
import traceback
from system import logger, config, workflow

def _workflow_classes(filename):
    """Returns a list of all subclasses of Workflow in a given module"""
    workflow_classes = []

    dirname, basename = os.path.split(filename)
    modulename, ext = os.path.splitext(basename)
    
    if dirname not in sys.path:
        sys.path.append(dirname)

    module = __import__(modulename)
    reload(module)
        
    for wfname in dir(module):
        wf = getattr(module, wfname)
        try:
            if issubclass(wf, workflow.Workflow):
                workflow_classes.append(wf)
        except TypeError, e:
            pass
    return workflow_classes
        
def workflow_list():
    """Returns a list containing all new workflows"""
    retval = []

    # List all .py files that can contain workflow classes
    wf_path = sys.modules['workflows'].__path__ + config.user_workflow_paths
    wf_files = []

    for dir in wf_path:
        if os.path.exists(dir) and os.path.isdir(dir):
            for fn in os.listdir(dir):
                if fn.endswith('.py') and ('#' not in fn):
                    wf_files.append(os.path.join(dir, fn))
        else:
            logger.log('warning', "Couldn't load workflows from '%s', directory not found (check preferences)." % dir)

    # Try to load each file and look for Workflow derived classes
    for fn in wf_files:
        try:
            for wf in _workflow_classes(fn):
                retval.append(wf)
        except Exception, e:
            logger.log('warning', 'Cannot load workflow: %s' % fn)
            logger.log('warning', e)

    return retval
