# Copyright(C), Endre Andersen, The SciCraft Team, www.scicraft.org
#Last update: 2/6-06


# $$$ ## This file is part of SciScraft.
# $$$ ##
# $$$ ## SciScraft is free software; you can redistribute it and/or modify it
# $$$ ## under the terms of the GNU General Public License as published by
# $$$ ## the Free Software Foundation; either version 2, or (at your option)
# $$$ ## any later version.
# $$$ ##
# $$$ ## SciScraft is distributed in the hope that it will be useful, but
# $$$ ## WITHOUT ANY WARANTY; without even the implied warranty of
# $$$ ## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# $$$ ## General Public License for more details.
# $$$ ##
# $$$ ## You should have received a copy of the GNU General Public License
# $$$ ## along with R; see the file COPYING.  If not, write to the Free
# $$$ ## Software Foundation, 59 Temple Place - Suite 330, Boston, MA
# $$$ ## 02111-1307, USA.
# $$$ 
# $$$ ## -*- texinfo -*-

# issues
# Modified T scores when using loadings
#	Use weights instead of loadings
#	Make sure weights are scaled Fixed
#	allow different pls methods Fixed
# Computing levrage values for samples DONE
# Computing residual vectors for samples and variables DONE
# Computing residual matrixes of X (Y is in fit) DONE
# Set up return data structure Fixed



PredictPLS <- function(fit,Xnew,NumOfComp){ 
  if(NumOfComp != -1) {
    if(NumOfComp < -1) {
      warning("wrong input NumOfComp should be larger than 0 reset to
  model optimum. ")
    } 
    else if(NumOfComp == 0) {
      warning("forced number of components to 0. Reset to model$ncomp. ")
    }
    else if(fit$ncomp < NumOfComp) {
      warning("Forced number of components, must be smaller than
               number of components in model. Set to model$ncomp. ")    }
    else {
      fit$ncomp <- NumOfComp
    }
  } 
  library(pls) 
  Xnew <- as.matrix(Xnew)
  scores <- predict.mvr(fit,list(X=Xnew),type = "scores")
  response <- predict.mvr(fit,list(X=Xnew),cumulative = FALSE,type = "response") 
  return(list(scores=as.matrix(scores),yPred=as.matrix(response)))  
} 



EAcv <- function (X, Y, ncomp,
        method = c("kernelpls", "simpls", "oscorespls","svdpc"),
        segments = 10,
        segment.type = c("random", "consecutive","interleaved"),
        length.seg, trace = FALSE, ...) {
    Y <- as.matrix(Y)
    dy <- dim(Y)
    dx <- dim(X)
    dnX <- dimnames(X)
    dnY <- dimnames(Y)
    dimnames(X) <- dimnames(Y) <- NULL
    if (is.list(segments)) {
        if (is.null(attr(segments, "type"))) 
            attr(segments, "type") <- "user supplied"
    }
    else {
      if (missing(length.seg)) {
        segments <- cvsegments(dx[1], k = segments, type = segment.type)
      }
      else {
        segments <- cvsegments(dx[1], length.seg = length.seg, 
                               type = segment.type)
      }
    }
    method <- match.arg(method)
    fitFunc <- switch(method, simpls = simpls.fit,
                kernelpls = kernelpls.fit,oscorespls = oscorespls.fit, svdpc = svdpc.fit)
    adj <- numeric(ncomp)
    cvPred <- pred <- array(dim = c(dx[1], dy[2], ncomp))
    
    #Inserted jack knife storage space here
    jkT <- array(NA,dim=c(dx[1],ncomp,length(segments)))
    jkW <- array(NA,dim=c(dx[2],ncomp,length(segments)))
    jkB <- array(NA,dim=c(dx[2],dy[2],ncomp,length(segments)))
   
    if (trace) 
        cat("Segment: ")
    for (n.seg in 1:length(segments)) {
      if (trace) 
        cat(n.seg, "")
      seg <- segments[[n.seg]]
      fit <- fitFunc(X[-seg, ], Y[-seg, ], ncomp, stripped = FALSE,...)
      #Insert jacknifing storage here
      #Store CV scores
      jkT[-seg,,n.seg] <- fit$scores 
      #Store CV loading weights
      jkW[,,n.seg] <- fit$loadings
      #Store CV Beta
      jkB[,,,n.seg] <- fit$coefficients
        
        
      Xtest <- sweep(X, 2, fit$Xmeans)
      for (a in 1:ncomp) pred[, , a] <- sweep(Xtest %*% fit$coefficients[, 
                                                                         , a], 2, fit$Ymeans, "+")
      cvPred[seg, , ] <- pred[seg, , , drop = FALSE]
      adj <- adj + length(seg) * colSums((pred - c(Y))^2)
    }
    if (trace) 
      cat("\n")
    MSEP0 <- apply(Y, 2, var) * dx[1]/(dx[1] - 1)
    MSEP <- colMeans((cvPred - c(Y))^2)
    R2 <- matrix(nrow = dy[2], ncol = ncomp)
    for (i in 1:dy[2]) R2[i, ] <- cor(cvPred[, i, ], Y[, i])^2
    objnames <- dnX[[1]]
    if (is.null(objnames)) 
      objnames <- dnY[[1]]
    yvarnames <- dnY[[2]]
    nCompnames <- paste(1:ncomp, "comps")
    names(MSEP0) <- yvarnames
    dimnames(adj) <- dimnames(MSEP) <- dimnames(R2) <- list(yvarnames,nCompnames)
    dimnames(cvPred) <- list(objnames, yvarnames, nCompnames)
    
    #add dimnames to jackknife results
    dimnames(jkT)<-list(objnames,nCompnames,as.character(segments))
    if (is.null(dnX[[2]])){
     varnames <- as.character(1:dx[2])}
    else{
     varnames <- dnX[[2]]
     }
    #print(varnames) 
    dimnames(jkW)<-list(varnames,nCompnames,as.character(segments))
    dimnames(jkB)<-list(varnames,yvarnames,nCompnames,as.character(segments))
    
    
    jkRes <- list(scores = jkT, loadings = jkW, coefficients= jkB)
    
    cvRes <-list(method = "CV", pred = cvPred, MSEP0 = MSEP0, MSEP = MSEP, 
                 adj = adj/dx[1]^2, R2 = R2, segments = segments)
    
    return(list(cvRes=cvRes,jkRes=jkRes))
        
}

jkPLS <-function(X,Y,fit,jk,NumberOfComponents,JackKnife,ModifiedTtest,penalty=0){
  
  ValidationSegments <- length(fit$validation$segments)
  aOpt <- NumberOfComponents


  dx <- dim(X)
  dy <- dim(Y)
 
  for (i in 1:ValidationSegments){
   #rotate scores
   rot <- procrustes(as.matrix(fit$scores[-fit$validation$segments[[i]],1:aOpt]),as.matrix(jk$jkRes$scores[-fit$validation$segments[[i]],1:aOpt,i]),scale = FALSE)
   jk$jkRes$scores[-fit$validation$segments[[i]],1:aOpt,i] <- rot$Yrot
   jk$jkRes$scores[fit$validation$segments[[i]],1:aOpt,i]  <- fit$scores[fit$validation$segments[[i]],1:aOpt]
   
   #rotate loadings using scores results
   jk$jkRes$loadings[,1:aOpt,i] <- jk$jkRes$loadings[,1:aOpt,i] %*% rot$rotation
 }
 
 #compute T statistics
  if (JackKnife=="T-scores on Coefficients"){
   t2 <- matrix(0,dy[2],dx[2])
   Sb <- matrix(0,dy[2],dx[2])
   
   #for each response variable
   
   Sbmod <- 0
   
   for (i in 1:dy[2]){
    Sb[i,] <- (diag(var(t(as.matrix(jk$jkRes$coefficients[,i,aOpt,])))))   
    dummy <- sort(Sb[i,])
    if (is.null(penalty)) {
     Sbmod <- median(Sb[i,])
     penalty <- Sbmod
    }
    else{
    Sbmod <- penalty
    }
    Sb[i,] <- (1-ModifiedTtest)*Sb[i,] + ModifiedTtest*Sbmod    
    
   t2[i,] <- ValidationSegments*fit$coefficients[,i,aOpt]/Sb[i,]
   }
   
  } 
  
  if (JackKnife=="T^2-scores on loadings"){
    t2 <- vector("numeric",length=dx[2])
    #Problem inflating d without centering it causes distance from origo to increase giving larger T scores upon inflation
    cvmx <- array(data = NA,dim=c(dx[2],aOpt,aOpt))
    
    for (i in 1:dx[2]){
     cvmx[i,,]<- cov(scale(t(as.matrix(jk$jkRes$loadings[i,1:aOpt,])), center = TRUE,scale=F))
    } 
    
    if (is.null(penalty)) {
     penalty <- apply(cvmx,c(2,3),median)
    }
      
    for (i in 1:dx[2]){
     cvmx[i,,] <- (1-ModifiedTtest)*cvmx[i,,] + ModifiedTtest*penalty  
    }
 
    hotell <- function(x,mx,ValidationSegments){
      x <- as.matrix(x)
      d <- mx
      tsq <- (ValidationSegments/(ValidationSegments-1))*t(d)%*%(ginv(x)%*%(d))
    }
    
    
    for (i in 1:dx[2]) {
     t2[i] <-hotell(cvmx[i,,],fit$loadings[i,1:aOpt],ValidationSegments)
    }
    
    t2 <- as.matrix(t2)
  }
  t2 <- list(t2=t2,penalty=penalty)
  return(t2)
  
}


 
 
############################################################################# 
##############################################################################
###############Main pls function##############################################
##############################################################################
##############################################################################

PLS <- function(X,Y,NumberOfComponents,Validation="none",ValidationSegments=10,JackKnife="none",ForceComponents="Use NumberOfComponents",FDR="none",ModifiedTtest=0,permutations= 50,method="oscorespls"){
  library(pls)
#  library(vegan)
  library(MASS)


if(!any(JackKnife == c("none","T-scores on Coefficients","T^2-scores on loadings"))){
 stop("invalid JackKnife valuse must be one of: none,T-scores on Coefficients,T^2-scores on loadings") 
}

if(!any(ForceComponents == c("Use NumberOfComponents","Find optimum"))){
 stop("invalid ForceComponents, valuse must be one of:Use NumberOfComponents,Find optimum") 
}

if(!any(Validation == c("none","Leave one out (LOO)","Random segments"))){
 stop("invalid Validation, valuse must be one of: none, Leave one out (LOO), Random segments") 
}

if(!any(FDR == c("none", "Compute FDR q-values"))){
 stop("Invalid FDR parameter, must be one of :none, Compute FDR q-values")
} 

if(!any(method==c("kernelpls", "simpls", "oscorespls"))){
 stop("Invalid method, method must be one of kernelpls, simpls, oscorespls")
}

#JackKnife("none","T-scores on Coefficients","T^2-scores on loadings") 
#Validation("none","Leave one out (LOO)","Random segments")
#ForceComponents("Use NumberOfComponents","Find optimum")
#FDR("none", "Compute FDR q-values")
##########################################################
###########Ensure exisence################################
  nperm <- permutations
  ModifiedTtest <- ModifiedTtest/100

  t2 <- 0
  qvalue <- 0
  Ypred <- 0

##########################################################
#########Quality control##################################


#check dims against NumberOfComponents and ValidationSegments
  X <- as.matrix(X)
  Y <- as.matrix(Y)
  dy <- dim(Y)
  dx <- dim(X)

  maxNcomp=min(dx)-1
  
  if (ValidationSegments > dx[1]){
   stop("Too many validation segments")
  }
  
  
  if (Validation =="none" & ForceComponents == "Find optimum"){
    stop("Can not find optimum number of components without validation")
  }
  if(NumberOfComponents > maxNcomp){
    NumberOfComponents <- maxNcomp
    warning("NumberofComponents too large reduced to min(dim(X))")
  }

  if (JackKnife == "T^2-scores on loadings" & (ValidationSegments < NumberOfComponents)){
    stop("Need more validation segments than components for q value calculations")
  }
  if (ModifiedTtest >= 1){
    stop("ModifiedTtest { [0,1>")}
  if (FDR=="Compute FDR q-values" & JackKnife == "none"){
    stop("FDR calculation requaires Jackknifing")
  }

  if (FDR=="Compute FDR q-values" & permutations == 0){
      stop("FDR calculation requires permutations e.g 50")
  }
  
  
  if(dx[1] != dy[1]){stop("need same number of samples in X and Y")}

  if (Validation=="none" & JackKnife !="none"){
    stop("Jackknifing requires some form of cross validation")}

  if (Validation == "none"){
    fit <- mvr(Y ~ X,data = list(X,Y),validation="none",ncomp=NumberOfComponents)
  }
  else if ((Validation != "none") &  (JackKnife == "none")){
    if(Validation == "Leave one out (LOO)"){
      fit <- mvr(Y ~ X,data = list(X,Y),validation="LOO",ncomp=NumberOfComponents,method=method)}
    if(Validation == "Random segments"){
      
      fit <- mvr(Y ~ X,data = list(X,Y),validation="CV",segments=ValidationSegments,ncomp=NumberOfComponents,method=method)}   
  }
  else{
    if(Validation=="Leave one out (LOO)"){ValidationSegments <- dx[1]} 
    jk <- EAcv(X,Y,ncomp=NumberOfComponents,segments=ValidationSegments,method=method)
    fit <- mvr(Y ~ X,data =list(X,Y),ncomp=NumberOfComponents,validation="none",method=method)
    fit$validation <- jk$cvRes
 }
 
  
 #Determine optimum model
  if ((Validation != "none") & (ForceComponents == "Find optimum")){
      if (is.matrix(fit$validation$MSEP)){
       
       tommy <- colSums(fit$validation$MSEP)}
      else{
      tommy <- fit$validation$MSEP}
      aOpt <- which.min(tommy)
  } 
    
  if (ForceComponents == "Use NumberOfComponents"){
    aOpt <- NumberOfComponents
  }

 
 # validate aOpt it still valid 
 if (dim(fit$scores)[2]<aOpt){
     aOpt <- dim(fit$scores)[2]   
  }
  
  if (JackKnife != "none"){ 
    t2 <- jkPLS(X,Y,fit,jk,NumberOfComponents=aOpt,JackKnife,ModifiedTtest,penalty=NULL)
    penalty <- t2$penalty
    t2 <- t2$t2
    
  }
 
  if (FDR=="Compute FDR q-values" & JackKnife != "none"){
    Tarray <- array(0,dim =c(dim(t2)[1],dim(t2)[2],nperm))
    
    for (i in 1:nperm){
      Ys <- sample(1:dy[1])
      Ys <- as.matrix(Y[Ys,])
      sjk <- EAcv(X,Ys,ncomp=NumberOfComponents,segments=ValidationSegments,method=method)
      sfit <- mvr(Ys ~ X,data =list(X,Y),ncomp=NumberOfComponents,validation="none",method=method)
      sfit$validation <- sjk$cvRes
      pertRes <- jkPLS(X,Ys,sfit,sjk,NumberOfComponents=aOpt,JackKnife,ModifiedTtest,penalty=NULL)  
      Tarray[,,i] <- pertRes$t2
      #print("one down")
    }
 
    #mediant <- apply(Tarray,c(1,2),median)
    qvalue  <- as.matrix(t2)
 
    if (JackKnife=="T-scores on Coefficients"){
      for (i in 1:dy[2]){
        for (j in 1:dx[2]) {
         n_better <- vector("numeric",length=nperm)
         for (k in 1:nperm){ 
          qv <- qvalue[i,j] 
          if (qv > 0) {    
            n_better[k] <- sum(Tarray[,j,k] >= qv)
          }
          if(qv < 0){
            n_better[k] <- sum(Tarray[,j,k] <= qv)
          }
         }
         
         if (qv > 0){
          qvalue[i,j] <- median(n_better)/sum(Tarray[,j,]>=qv)
         }
         
         if (qv > 0){
          qvalue[i,j] <- median(n_better)/sum(Tarray[,j,]<=qv)
         }
         
         
        } 
      }
    }
    
    if(JackKnife == "T^2-scores on loadings"){
      for (i in 1:dx[2]){
       n_better <- vector("numeric",nperm)
       qv <- t2[i,1]
       for (j in 1:nperm){
        n_better[j] <- sum(Tarray[,,j] >= qv) 
       }
        qvalue[i,1] <- median(n_better)/(sum(t2>=qv)) #mental noe how is fdr computed
      }
    } 
  
    qvalue[qvalue > 1] <- 1
  }
#compute q values
    
  Xmeans <- X

  for (i in 1:dx[1]){
    Xmeans[i,] <- fit$Xmeans
  }
  Xres <- X - Xmeans - fit$scores[,1:aOpt]%*%t(fit$loadings[,1:aOpt])  
  XVarRes <- colSums(Xres)  
  XSamRes <- rowSums(Xres)
  Leverage <- vector("numeric",length= dx[1])  
  for (i in 1:dx[1]){
    Leverage[i] <-100*((1/dx[1]) + (t(fit$scores[i,1:aOpt])%*%ginv(t(fit$scores[,1:aOpt])%*%(fit$scores[,1:aOpt]))%*%(fit$scores[i,1:aOpt])))
  }

#additional output:
#
  if (Validation!="none"){
    diagnostics <- list(optimumNumberOfComponents = aOpt,Tstat= t2,qValue = qvalue, Yval=fit$validation$pred[,,], Ypred = as.matrix(fit$validation$pred[,,aOpt]),E=as.matrix(Xres),Leverage=as.matrix(Leverage),XVarRes=as.matrix(XVarRes),XSamRes=XSamRes)
  }
  else {
    diagnostics <- list(optimumNumberOfComponents = aOpt,Tstat= 0,qValue = 0, Yval=fit$validation$pred[,,], Ypred = as.matrix(fit$fitted[,,aOpt]),E=Xres,Leverage=as.matrix(Leverage),XVarRes=as.matrix(XVarRes),XSamRes=as.matrix(XSamRes))
  }
  #convert explained variance to percent.
  fit$Xvar <- explvar(fit) 

  
  return(list(plsModel=fit,diagnostics = diagnostics))#, t2 = t2, qval = qvalue,mt = mediant))
}
