DistanceMatrix<-function(X,metric = "euclidean", p=2,rows="rows"){
#compute distance measures  
#metric "euclidean", "maximum", "manhattan", "canberra","binary","minkowski" "pearson correlation","kendall correlation"
#"spearman correlation"

#Last update: 26/3-06
# $$$ ## This file is part of SciCraft.
# $$$ ##
# $$$ ## SciCraft is free software; you can redistribute it and/or modify it
# $$$ ## under the terms of the GNU General Public License as published by
# $$$ ## the Free Software Foundation; either version 2, or (at your option)
# $$$ ## any later version.
# $$$ ##
# $$$ ## SciCraft is distributed in the hope that it will be useful, but
# $$$ ## WITHOUT ANY WARRANTY; without even the implied warranty of
# $$$ ## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# $$$ ## General Public License for more details.
# $$$ ##
# $$$ ## You should have received a copy of the GNU General Public License
# $$$ ## along with R; see the file COPYING.  If not, write to the Free
# $$$ ## Software Foundation, 59 Temple Place - Suite 330, Boston, MA
# $$$ ## 02111-1307, USA.
# $$$ 
# $$$ ## -*- texinfo -*-
#
#  
  library(cluster)
  if (rows=="columns"){
    X = t(X)
  }


#bicov and bicor taken from
#http://www.unt.edu/benchmarks/archives/2001/december01/rss.htm

  bicov<-function(x,y){
    Xna_ind <- is.na(x)
    Yna_ind <- is.na(y)
    use_ind <- !(Xna_ind | Yna_ind)
    x<-x[use_ind]
    y<-y[use_ind]
    
    mx <- median(x)
    my <- median(y)
    ux <- abs((x - mx)/(9 * qnorm(0.75) * mad(x)))
    uy <- abs((y - my)/(9 * qnorm(0.75) * mad(y)))
    aval <- ifelse(ux <= 1, 1, 0)
    bval <- ifelse(uy <= 1, 1, 0)
    top <- sum(aval * (x - mx) * (1 - ux^2)^2 * 
               bval * (y - my) * (1 - uy^2)^2)
    top <- length(x) * top
    botx <- sum(aval * (1 - ux^2) * (1 - 5 * ux^2))
    boty <- sum(bval * (1 - uy^2) * (1 - 5 * uy^2))
    bi <- top/(botx * boty)
    bi}
  
  bicor<-function(x,y){
    x<-as.numeric(x)
    y<-as.numeric(y)
    bicov(x,y)/(sqrt(bicov(x,x)*bicov(y,y)))}
  
  
  bootcor <-function(x,y,samples=1000,method = "robust"){
    cc <- vector(mode="numeric",length=samples)
    rc <- cc
    
    for (i in 1:samples){
      ind<-sample(1:length(x),replace=T)
      cc[i]<- cor(x[ind],y[ind],use="pairwise.complete.obs")
      rc[i] <- bicor(x[ind],y[ind])
    }
    
    if(method=="robust"){
      co <- median(rc)}
    else{co <- median(cc)}
    
    co
  } 
  
  
  if (any(metric == c("euclidean", "maximum", "manhattan", "canberra","binary","minkowski"))){
    D <-as.matrix(dist(X, method = metric, diag = FALSE, upper = TRUE, p = p))
  }
  if (metric == "pearson correlation"){
    X <- t(X)
    D <- cor(X,use="pairwise.complete.obs")
    D <- 1-D
  }
  if (metric == "kendall correlation"){
    X <- t(X)
    D <- cor(X,method="kendall",use="pairwise.complete.obs")
    D <- 1-D
  }
  
  if (metric == "spearman correlation"){
    X <- t(X)
    D <- cor(X,method="spearman", use="pairwise.complete.obs")
    D <- 1-D
  }
  
  if (metric == "bicor"){
    D <-matrix(0,nrow=dim(X)[1],ncol=dim(X)[1])
    for (i in 1:dim(X)[1]){
      for (j in i:dim(X)[1]){
        if (i==j){D[i,j] = 1}
        else{
          D[i,j] <- bicor(X[i,],X[j,]) 
          D[j,i] <- D[i,j]}
        
      }
    }
    
    D <- 1-D
  }
  
  
  if (metric=="bootstrapCor"){
    D <-matrix(0,nrow=dim(X)[1],ncol=dim(X)[1])
    for (i in 1:dim(X)[1]){
      for (j in i:dim(X)[1]){
        if (i==j){D[i,j] = 1}
        else{
          D[i,j] <- bootcor(X[i,],X[j,],method="cor") 
          D[j,i] <- D[i,j]}
        
      }
    }
    D <- 1-D
  }
  
  
  if (metric=="bootstrapBiCor"){
    D<-matrix(0,nrow=dim(X)[1],ncol=dim(X)[1])
    for (i in 1:dim(X)[1]){
      for (j in i:dim(X)[1]){
        if (i==j){D[i,j] = 1}
        else{
          D[i,j] <- bootcor(X[i,],X[j,],method="robust") 
          D[j,i] <- D[i,j]}
        
      }
    }
    D <- 1-D
  }
  
  return (list(DistanceMatrix = D))
  
}
